#########################
# Zope Army Knife       #
# (c) 2003,2004 NAUMEN  #
#########################
"""
Revision information:
$Id: Watch.py,v 1.2 2004/04/06 17:21:20 kvas Exp $

ZAK class
"""
__version__ = '$Revision: 1.2 $'[11:-2]

from __future__ import nested_scopes

from threading import Lock
from zLOG import LOG, WARNING
from traceback import extract_stack, format_list
from threading import currentThread
from time import strftime
from Persistence import Persistent


class WatchRecord(Persistent):
    """Watch record in ZODB. Makes watch persistent"""

    __roles__ = None
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, method_path):
        """creates new watch and attaches it to specified class"""
        self.__method_path = method_path
        self.__pre_script_src = ''
        self.__post_script_src = ''
        self.activate()

    def activate(self):
        if not self.isActive():
            w = ActiveWatches.watch(self.__method_path)
            w.setPreScript(self.__pre_script_src)
            w.setPostScript(self.__post_script_src)

    def deactivate(self):
        if self.isActive():
            ActiveWatches.unwatch(self.__method_path)

    def isActive(self):
        return ActiveWatches.has(self.__method_path)

    def __getWatch(self):
        return ActiveWatches.get(self.__method_path)

    def setPreScript(self, script):
        self.__pre_script_src = script
        if self.isActive():
            self.__getWatch().setPreScript(script)

    def setPostScript(self, script):
        self.__post_script_src = script
        if self.isActive():
            self.__getWatch().setPostScript(script)

    def getMethodPath(self):
        """return class pathname + method name"""
        return self.__method_path

    def getPreScript(self):
        return self.__pre_script_src

    def getPostScript(self):
        return self.__post_script_src

    def getCR(self, index):
        """Return method call record"""
        if self.isActive():
            return self.__getWatch().call_records[int(index)]
        else:
            return None

    def getNCalls(self):
        """return the number of watched calls"""
        if self.isActive():
            return len(self.__getWatch().call_records)
        else:
            return '-'

    def listCalls(self):
        """return a list of the calls made"""
        if self.isActive():
            return self.__getWatch().call_records
        else:
            return []
        
    def clear(self):
        """Forget all call records"""
        if self.isActive():
            self.__getWatch().call_records = []


class Watch:
    """Method invocation watch
    
    Intercepts calls to a specific method and logs them
    """

    def __init__(self, method_path):
        """creates new watch and attaches it to specified class"""
        self.method_path = method_path
        self.call_records = []
        self.__attached = 0
        self.__pre_script = None
        self.__post_script = None

    def addCR(self, cr):
        """Log method call"""
        self.call_records.append(cr)

    def setPreScript(self, script):
        self.__pre_script = compile(script, 'pre script', 'exec')

    def runPreScript(self, context):
        if self.__pre_script:
            eval(self.__pre_script, context)

    def setPostScript(self, script):
        self.__post_script = compile(script, 'pre script', 'exec')

    def runPostScript(self, context):
        if self.__post_script:
            eval(self.__post_script, context)

    def createContext(self, args, kwargs):
        return { 'args': args,
                 'kwargs': kwargs,
                 'run_method': 1,
                 'ret_val': None }

    def makeWrapper(self, wrappee, docstring=None):
        """Create a wrapper around wrappee"""
        code = compile("""def wrapper(*args, **kwargs):
            try:
                cr = CallRecord(extract_stack(), args, kwargs, currentThread())
            except TypeError:  # if ZAK was refreshed, all will fail
                return wrappee(*args, **kwargs)

            context = watch.createContext(args, kwargs)
            try:
                watch.runPreScript(context)
                if context['run_method']:
                    context['ret_val'] = wrappee(*context['args'],
                                                 **context['kwargs'])
                watch.runPostScript(context)
            finally:
                cr.storeRetVal(context['ret_val'])
                watch.addCR(cr)
            return context['ret_val']\n""", 'makeWrapper_virtual', 'exec')
        g = { 'CallRecord': CallRecord,
              'extract_stack': extract_stack,
              'currentThread': currentThread }
        eval(code, g)
        wrapper = g['wrapper']
        wrapper.func_globals['watch'] = self
        wrapper.func_globals['wrappee'] = wrappee
        if docstring is not None:
            wrapper.__doc__ = docstring
        return wrapper
        
    def attach(self):
        """attach self to specified class"""
        if self.__attached: return

        t = self.method_path.split('.')
        try:                               # try to import as a module function
            module = '.'.join(t[:-1])
            func_name = t[-1]
            module = __import__(module, globals(), locals(), ['zoo'])
            func = unwrap(getattr(module, func_name)) # unwrap it...just in case
            setattr(module, func_name, self.makeWrapper(func))
            self.__type = 'module'
            self.__func_name = func_name
            self.__module = module
        except:                             # try as a class method
            module = '.'.join(t[:-2])
            class_name = t[-2]
            method_name = t[-1]
            module = __import__(module, globals(), locals(), ['zoo'])
            klass = getattr(module, class_name)
            method = unwrap(getattr(klass, method_name).im_func)
            setattr(klass, method_name,
                    self.makeWrapper(method, getattr(method, '__doc__', None)))
            self.__type = 'class'
            self.__method_name = method_name
            self.__class = klass
            
        self.__attached = 1

    def detach(self):
        """detach self from target"""
        if not self.__attached: return

        try:
            if self.__type == 'module':
                wrapper = getattr(self.__module, self.__func_name)
                setattr(self.__module, self.__func_name,
                        wrapper.func_globals['wrappee'])
            elif self.__type == 'class':
                wrapper = getattr(self.__class, self.__method_name).im_func
                setattr(self.__class, self.__method_name,
                        wrapper.func_globals['wrappee'])
        except:
            LOG('ZAK', WARNING,
                'detaching watch on %s failed (sometimes it is ok)' % \
                                                   self.getMethodPath())

        self.__attached = 0


def alreadyWrapped(func):
    try:
        wp = func.func_globals['wrappee']
        if func.func_name == 'wrapper': return 1
    except KeyError:
        pass
    return 0
    
def unwrap(func):
    while alreadyWrapped(func):
        func = func.func_globals['wrappee']
    return func


class WatchContainer:
    """Synchronized watch container"""
    def __init__(self):
        self.__lock = Lock()
        self.__watches = {}

    def watch(self, mp):
        self.__lock.acquire()
        try:
            watch = Watch(mp)
            if self.__watches.has_key(mp):
                raise ValueError, 'Already watching ' + mp
            watch.attach()
            self.__watches[mp] = watch
            return watch
        finally:
            self.__lock.release()

    def unwatch(self, mp):
        self.__lock.acquire()
        try:
            self.__watches[mp].detach()
            del self.__watches[mp]
        finally:
            self.__lock.release()

    def get(self, mp):
        return self.__watches[mp]

    def has(self, mp):
        return self.__watches.has_key(mp)

    def items(self):
        return self.__watches.items()


ActiveWatches = WatchContainer()


def html_quote(s):
    """escape <, > and &"""
    return s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')

class CallRecord:
    """watch log record"""
    def __init__(self, tb, args, kwargs, thread):
        self.time = strftime("%H.%M.%S %d/%m/%Y")
        self.tb = '<br>'.join(map(html_quote, format_list(tb)[:-1]))
        self.args = ',\n'.join(map(repr, args))
        self.kwargs = repr(kwargs)
        self.thr_id = thread.getName()

    def storeRetVal(self, ret_val):
        self.ret_val = repr(ret_val)
